/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <poll.h>
#include <stdio.h>
#include <stdint.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "network.h"
#include "date.h"
#include "block.h"
#include "protocol.h"
#include "common.h"
#include "config.h"
#include "array.h"
#include "buffer.h"
#include "exec.h"
#include "units.h"
#include "client.h"

// **************************************************************************
// 
// **************************************************************************
#define DEBUG(...)	\
	do { \
		fprintf(stderr, "%s ", curdate_str()); \
		fprintf(stderr, __VA_ARGS__); \
	} while(0)

typedef struct client_t {
	int fd;
	
	buffer inbuff;
	buffer outbuff;

	unit_list units;
	int state;
	
	array pfd_a;
} client;


// **************************************************************************
// 
// **************************************************************************
static int client_io(client *cl)
{
	struct pollfd *pfd;
	unit *u;
	short inev;
	int i;
	
	// poll
	pfd = (struct pollfd *)array_get(&cl->pfd_a, 1 + (cl->units.size * 3));
	
	pfd[0].fd = cl->fd;
	pfd[0].events = POLLIN;
	if (buffer_io_complete(&cl->outbuff)) {
		buffer_io_reset(&cl->outbuff);
		buffer_reset(&cl->outbuff);
		inev = POLLIN;
	} else {
		// we refuse to receive any data from the tasks' stdout
		// and stderr until the socket's output buffer is empty
		inev = 0;
		pfd[0].events |= POLLOUT;
	}
	pfd[0].revents = 0;
	
	u = cl->units.head;
	for (i = 0; i < cl->units.size; i++) {
		// *output* to unit's stdin
		if (buffer_io_complete(&u->inbuff)) {
			if (u->exe.infd >= 0) {
				DEBUG("[%d]: closing stdin\n", u->id);
				close(u->exe.infd);
				u->exe.infd = -1;
			}
			pfd[1 + (i * 3) + 0].fd = -1;
			pfd[1 + (i * 3) + 0].events = 0;
			pfd[1 + (i * 3) + 0].revents = 0;
		} else {
			pfd[1 + (i * 3) + 0].fd = u->exe.infd;
			pfd[1 + (i * 3) + 0].events = POLLOUT;
			pfd[1 + (i * 3) + 0].revents = 0;
		}
		
		// *input* from unit's stdout
		pfd[1 + (i * 3) + 1].fd = u->exe.outfd;
		pfd[1 + (i * 3) + 1].events = inev;
		pfd[1 + (i * 3) + 1].revents = 0;

		// *input* from unit's stderr
		pfd[1 + (i * 3) + 2].fd = u->exe.errfd;
		pfd[1 + (i * 3) + 2].events = inev;
		pfd[1 + (i * 3) + 2].revents = 0;
		
		u = u->next;
	}

	if (poll(pfd, 1 + (cl->units.size * 3), -1) < 0) {
		fprintf(stderr, "client_io(): poll(): %s\n", strerror(errno));
		return(-1);
	}

	if (pfd[0].revents & (POLLIN|POLLERR|POLLHUP))
		buffer_io_read(&cl->inbuff, cl->fd);
	
	if (pfd[0].revents & (POLLOUT|POLLERR|POLLHUP))
		buffer_io_write(&cl->outbuff, cl->fd);
	
	if ((cl->inbuff.eof) || (cl->inbuff.error)
	||  (cl->outbuff.eof) || (cl->outbuff.error)) {
		if (cl->inbuff.eof)
			DEBUG("client_io(): connection closed (in)\n");
		if (cl->outbuff.eof)
			DEBUG("client_io(): connection closed (out)\n");
		if ((cl->inbuff.error) || (cl->outbuff.error))
			DEBUG("i/o error on server socket\n");
		return(-1);
	}
	
	u = cl->units.head;
	for (i = 0; i < cl->units.size; i++) {
		if (pfd[1 + (i * 3) + 0].revents & (POLLOUT|POLLERR|POLLHUP))
			buffer_io_write(&u->inbuff, u->exe.infd);
		
		if (pfd[1 + (i * 3) + 1].revents & (POLLIN|POLLERR|POLLHUP))
			buffer_io_read(&u->outbuff, u->exe.outfd);
		
		if (pfd[1 + (i * 3) + 2].revents & (POLLIN|POLLERR|POLLHUP))
			buffer_io_read(&u->errbuff, u->exe.errfd);
		
		if ((u->inbuff.eof) || (u->inbuff.error)
		||  ((u->outbuff.eof) && (u->errbuff.eof))
		||  (u->outbuff.error) || (u->errbuff.error)) {
			if ((u->inbuff.error) || (u->outbuff.error)
			||  (u->errbuff.error)) {
				DEBUG("[%d]: errors %d %d %d: %s\n",
					u->id, u->inbuff.error,
					u->outbuff.error, u->errbuff.error,
					strerror(errno));
			}
			
			DEBUG("[%d]: finished; totals %zd %zd %zd\n",
				u->id,
				u->inbuff.total_write, u->outbuff.total_read,
				u->errbuff.total_read);
			
			u->flags |= UNIT_END;
		}
		
		u = u->next;
	}
	
	return(0);
}

// **************************************************************************
// 
// **************************************************************************
static int client_sched_server(client *cl)
{
	int h, r;
	int tid;
	uint64_t ping_id;
	char *shell, *dir, *opts, *name, *cmd;
	unit *u;
	
	if (cl->outbuff.size > conf.proto_buff_max) {
		DEBUG("output buffer too large (%zd > %zd)\n",
			cl->outbuff.size, conf.proto_buff_max);
		return(-1);
	}
	
	if (proto_prepare(&cl->inbuff, PROTO_SV_MAGIC, conf.proto_buff_max)) {
		DEBUG("protocol error\n");
		return(-1);
	}

	if (!proto_ready(&cl->inbuff))
		return(0);
	
	h = proto_header(&cl->inbuff);
	
	// **********************************
	if (h == PROTO_SV_RUN) {
		if (proto_fetch(&cl->inbuff, "isssss",
				&tid, &shell, &dir, &opts, &name, &cmd)) {
			DEBUG("PROTO_SV_RUN: bad packet\n");
			return(-1);
		}
		
		u = unit_new(tid, shell, dir, opts, name, cmd);
		
		if (!u) {
			//proto_append(&cl->outbuff,
			//	PROTO_CL_MAGIC, PROTO_CL_END,
			//	"ii", tid, PROTO_END_FAIL);
			//return(0);
			DEBUG("PROTO_SV_RUN: unit_new() failed.\n");
			return(-1);
		}
		
		ul_append(&cl->units, u);
		
		DEBUG("[%d]: started\n", tid);
		return(0);
	}
	
	// **********************************
	if (h == PROTO_SV_STOP) {
		if (proto_fetch(&cl->inbuff, "ii", &tid, &r)) {
			DEBUG("PROTO_SV_STOP: bad packet\n");
			return(-1);
		}
		
		u = ul_find(&cl->units, tid, 0);
		if (!u) {
			DEBUG("PROTO_SV_STOP: task [%d] not found, "
				"dismissed\n", tid);
			return(0);
		}
		
		ul_unlink(&cl->units, u);
		unit_free(u);
		
		proto_append(&cl->outbuff, PROTO_CL_MAGIC, PROTO_CL_END,
			"ii", tid, r);
		
		DEBUG("[%d]: stopped\n", tid);
		return(0);
	}

	// **********************************
	if (h == PROTO_SV_PING) {
		if (proto_fetch(&cl->inbuff, "p", &ping_id)) {
			DEBUG("PROTO_SV_PING: bad packet\n");
			return(-1);
		}
		
		proto_append(&cl->outbuff, PROTO_CL_MAGIC, PROTO_CL_PONG,
			"p", ping_id);
		
		//DEBUG("[ping] %zd\n", ping_id);
		return(0);
	}

	// **********************************
	DEBUG("unknown server header 0x%08x\n", h);
	return(-1);
}

// **************************************************************************
// 
// **************************************************************************
static void client_sched_unit(client *cl, unit *u)
{
	if (u->outbuff.partial) {
		proto_append(&cl->outbuff, PROTO_CL_MAGIC, PROTO_CL_OUT,
			"id", u->id, u->outbuff.data, u->outbuff.partial);
	
		buffer_io_reset(&u->outbuff);
	}
	
	if (u->errbuff.partial) {
		proto_append(&cl->outbuff, PROTO_CL_MAGIC, PROTO_CL_ERR,
			"id", u->id, u->errbuff.data, u->errbuff.partial);
	
		buffer_io_reset(&u->errbuff);
	}

	if (u->flags & UNIT_END) {
		proto_append(&cl->outbuff, PROTO_CL_MAGIC, PROTO_CL_END,
			"ii", u->id, PROTO_END_DONE);

		ul_unlink(&cl->units, u);
		unit_free(u);
		return;
	}
}


// **************************************************************************
// 
// **************************************************************************
/*
static int client_sched_watch(client *cl)
{
	unit *u, *unext;
	size_t o;
	char p;
	int state;
	
	// We handle only the watch unit
	u = cl->watch;
	if (!u)
		return(0);
	
	state = cl->state;
	
	if (u->outbuff.partial) {
		for (o = 0; o < u->outbuff.partial; o++) {
			p = ((char *)u->outbuff.data)[o];
			
			if (p == '0')
				state = 0;
			else if (p == '1')
				state = 1;
		}
		
		buffer_io_reset(&u->outbuff);
	}
	
	if (u->errbuff.partial) {
		buffer_io_reset(&u->errbuff);
	}

	if (u->flags & UNIT_END) {
		DEBUG("Watch i/o error\n");
		return(-1);
	}
	
	if (state == cl->state)
		return(0);
	
	// switch to on
	if (state == 1) {
		DEBUG("watch: switching to ON (with %d processes)\n",
			cl->cores);
		proto_append(&cl->outbuff, PROTO_CL_MAGIC, PROTO_CL_SETUP,
			"is", cl->cores, conf.host);
		cl->state = 1;
		return(0);
	}

	// switch to off, then interrupt tasks
	DEBUG("watch: switching to OFF\n");
	proto_append(&cl->outbuff, PROTO_CL_MAGIC, PROTO_CL_SETUP,
		"is", 0, conf.host);
	cl->state = 0;

	u = cl->units.head;
	while(u) {
		unext = u->next;
		
		if (!(u->flags & UNIT_WATCH)) {
			proto_append(&cl->outbuff, PROTO_CL_MAGIC,
				PROTO_CL_END, "ii", u->id, PROTO_END_FAIL);
		
			DEBUG("[%d]: interrupted by watch\n", u->id);

			ul_unlink(&cl->units, u);
			unit_free(u);
		}
		
		u = unext;
	}

	return(0);
}
*/
// **************************************************************************
// 
// **************************************************************************
static int client_sched(client *cl)
{
	unit *u, *unext;
	
	if (client_sched_server(cl))
		return(-1);
	
	u = cl->units.head;
	while (u) {
		unext = u->next;
		
		client_sched_unit(cl, u);
		
		u = unext;
	}
	
	return(0);
}

// **************************************************************************
// 
// **************************************************************************
int ben_client(char *addr, char *name, int processes, int flags)
{
	client cl;
	
	// init
	cl.fd = socket_connect(addr);
	
	if (cl.fd < 0)
		return(1);
	
	printf("%s %s\n", BEN_CLIENT_ACK_STR, addr);
	fflush(stdout);
	
	if (flags & BEN_FLAG_DAEMONIZE) {
		if (exec_daemonize(NULL, NULL, conf.client_log)) {
			close(cl.fd);
			return(1);
		}
	}
	
	buffer_init(&cl.inbuff, 0);
	buffer_init(&cl.outbuff, 0);
	ul_init(&cl.units);
	array_init(&cl.pfd_a, sizeof(struct pollfd), 0);

	// first setup
	cl.state = 1;
	proto_append(&cl.outbuff, PROTO_CL_MAGIC, PROTO_CL_SETUP,
		"is", processes, name);

	// loop
	while (1) {
		if (client_io(&cl))
			break;
		
		if (client_sched(&cl))
			break;
	}
	
	// cleanup
	close(cl.fd);
	buffer_clear(&cl.inbuff);
	buffer_clear(&cl.outbuff);
	ul_clear(&cl.units);
	array_clear(&cl.pfd_a);
	
	return(1);
}


