/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#define _POSIX_C_SOURCE 201605L
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include "xmem.h"
#include "block.h"
#include "exec.h"

#define DEBUG(...)	fprintf(stderr, __VA_ARGS__)

// **************************************************************************
//
// **************************************************************************
static void exec_close(int fd)
{
	if (fd < 0)
		return;
	
	if (close(fd)) {
		DEBUG("close(): %s\n", strerror(errno));
	}
}

// **************************************************************************
//
// **************************************************************************
int exec_fork(exec_unit *u, char **argv, char **add_envp, int flags)
{
	int p[6];
	int i, j;
	char **e;
	
	for (j = 0; j < 6; j++)
		p[j] = -1;

	for (i = 0; i <= 2; i++) {
		if (flags & (1 << i))
			continue;
		
		if (pipe(p + (i * 2))) {
			DEBUG("exec_fork(): pipe(%d): %s\n",
				i, strerror(errno));
			for (j = 0; j < 6; j++)
				exec_close(p[j]);
			return(-1);
		}
	}

	u->flags = flags;
	u->pid = fork();
	
	if (u->pid == -1) {
		DEBUG("exec_fork(): fork(): %s\n", strerror(errno));
		for (j = 0; j < 6; j++)
			exec_close(p[j]);
		return(-1);
	} else if (u->pid != 0) {
		// parent
		u->infd = p[1];
		u->outfd = p[2];
		u->errfd = p[4];
		exec_close(p[0]);
		exec_close(p[3]);
		exec_close(p[5]);
		return(0);
	}
	
	// child
	
	if (flags & EXEC_GROUP_LEADER) {
		// Set as process group leader, so that later we can kill
		// the entire process group easily. As a side-effect,
		// this puts the child in the background (if still attached
		// to a tty).
		if (setpgid(0, 0)) {
			DEBUG("setpgid(0, 0): %s\n", strerror(errno));
			exit(1);
		}
	}
	
	exec_close(p[1]);
	exec_close(p[2]);
	exec_close(p[4]);

	for (i = 0; i <= 2; i++) {
		j = (i == 0) ? 0 : (i * 2) + 1;
		
		if (p[j] < 0)
			continue;

		if (dup2(p[j], i) == -1) {
			DEBUG("dup2(p[%d] == %d, %d): %s\n",
				j, p[j], i, strerror(errno));
			exit(1);
		}
		
		if (p[j] > 2)
			exec_close(p[j]);
	}
	
	e = add_envp;
	if (e) {
		while (e[0]) {
			if (!e[1]) {
				DEBUG("%s: setenv(\"%s\", nil)\n",
					argv[0], e[0]);
				exit(1);
			}
			
			if (setenv(e[0], e[1], 1)) {
				DEBUG("%s: setenv(\"%s\", \"%s\", 1): %s\n",
					argv[0], e[0], e[1], strerror(errno));
				exit(1);
			}
			
			e += 2;
		}
	}
	
	execvp(argv[0], argv);
	
	DEBUG("%s: %s\n", argv[0], strerror(errno));
	
	exit(1);
}

// **************************************************************************
//
// **************************************************************************
int exec_terminate(exec_unit *u, int *r_status)
{
	pid_t wr, kpid;
	int status;
	
	if (r_status)
		*r_status = -1;
	
	if (u->pid < 0)
		return(0);
	
	exec_close(u->infd);
	exec_close(u->outfd);
	exec_close(u->errfd);

	u->infd = -1;
	u->outfd = -1;
	u->errfd = -1;
	
	// kill the entire process group
	kpid = (u->flags & EXEC_GROUP_LEADER) ? -u->pid : u->pid;
	
	if (kill(kpid, SIGTERM)) {
		DEBUG("kill(): %s\n", strerror(errno));
		errno = 0;
	}
	
	if (u->flags & EXEC_ORPHANED) {
		// Right now we can't get the status if we daemonize
		// because the exec'ed process is a sibling of the
		// daemon. This could be mitigated by exec'ing from
		// the daemon, but it would require something like a
		// exec_daemonize_delayed() / exec_daemonize_release()
		// pair.
		if (r_status)
			*r_status = 0;
		return(0);
	}
	
	// reap shell process (blocking wait)
	while (1) {
		wr = waitpid(u->pid, &status, 0);
		
		if (wr == -1) {
			if (errno != EINTR) {
				DEBUG("waitpid(): %s\n",
					strerror(errno));
				return(-1);
			}
		} else if (wr == u->pid) {
			u->pid = -1;
			
			if (r_status) {
				if (WIFEXITED(status))
					*r_status = WEXITSTATUS(status);
			}
			
			return(0);
		} else {
			DEBUG("waitpid(): returned pid %d != %d\n",
				wr, u->pid);
			return(-1);
		}
	}
}

// ***************************************************************************
// 
// ***************************************************************************
int exec_daemonize(const char *inf, const char *outf, const char *errf)
{
	const char *spath[3];
	int sfd[3], t, u;
	pid_t child;

	// redirect to provided path or /dev/null
	spath[0] = ( inf) ?  inf : "/dev/null";
	spath[1] = (outf) ? outf : "/dev/null";
	spath[2] = (errf) ? errf : "/dev/null";

	// open paths
	for (t = 0; t <= 2; t++) {
		if (!isatty(t)) {
			sfd[t] = -1;
			errno = 0;
			continue;
		}
		
		exec_path_mkdir(spath[t]);
		
		sfd[t] = open(spath[t],
			(t == 0) ? O_RDONLY : (O_CREAT | O_WRONLY | O_TRUNC),
			0600);

		if (sfd[t] < 0) {
			DEBUG("%s: %s\n", spath[t], strerror(errno));
			for (u = 0; u < t; u++)
				close(sfd[u]);
			return(-1);
		}
	}
	
	// fork
	child = fork();
	
	// fork error
	if (child < 0) {
		DEBUG("fork(): %s\n", strerror(errno));
		for (u = 0; u <= 2; u++)
			close(sfd[u]);
		return(-1);
	}
	
	// fork: parent exits with success code 0
	if (child > 0) {
		exit(0);
	}

	// fork: child goes on
	for (t = 0; t <= 2; t++) {
		if ((sfd[t] >= 0) && (sfd[t] != t)) {
			if (dup2(sfd[t], t) < 0) {
				DEBUG("dup2(%s, %d): %s\n",
					spath[t], t, strerror(errno));
				exit(1);
			}
			
			close(sfd[t]);
		}
	}
	
	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
int exec_signals_setup(void)
{
	struct sigaction siga;
	
	siga.sa_handler = SIG_IGN;
	sigemptyset(&siga.sa_mask);
	siga.sa_flags = 0;
	
	if (sigaction(SIGPIPE, &siga, NULL)) {
		DEBUG("sigaction(SIGPIPE): %s\n", strerror(errno));
		return -1;
	}
	
	return 0;
}

// ***************************************************************************
// 
// ***************************************************************************
#define EXEC_LAST_COMPONENT	1
static int exec_internal_mkdir(const char *path, int flags)
{
	char *buff;
	size_t len, max, t;
	
	len = strlen(path);
	buff = xmalloc(len + 1);
	memcpy(buff, path, len + 1);
	
	max = (flags & EXEC_LAST_COMPONENT) ? len : len - 1;
	
	for (t = 1; t <= max; t++) {
		if ((buff[t] == '/') || (buff[t] == '\0')) {
			buff[t] = 0;
			
			if (mkdir(buff, 0700)) {
				if (errno == EEXIST) {
					errno = 0;
				} else {
					DEBUG("mkdir(%s): %s: %s\n",
						path, buff, strerror(errno));
					xfree(buff);
					return(-1);
				}
			}
			
			buff[t] = '/';
		}
	}
	
	xfree(buff);
	
	return(0);
}


int exec_recursive_mkdir(const char *path)
{
	return(exec_internal_mkdir(path, EXEC_LAST_COMPONENT));
}

int exec_path_mkdir(const char *path)
{
	return(exec_internal_mkdir(path, 0));
}


// ***************************************************************************
// 
// ***************************************************************************
int exec_mkchdir(const char *path)
{
	if (chdir(path) == 0)
		return(0);
	
	if (errno == ENOENT) {
		if (mkdir(path, 0700)) {
			DEBUG("mkdir(): %s: %s\n",
				path, strerror(errno));
			return(-1);
		}
		
		if (chdir(path) == 0)
			return(0);
	}
	
	DEBUG("chdir(): %s: %s\n", path, strerror(errno));
	return(-1);
}

// **************************************************************************
//
// **************************************************************************
int exec_pager_run(exec_pager *pg)
{
	int p[2], ofd;
	char *pager, *argv[2];
	pid_t pid;
	
	pg->pid = 0;
	pg->old_out = -1;

	if (!isatty(1)) {
		errno = 0;
		return(0);
	}
	
	if (pipe(p)) {
		DEBUG("exec_pager(): pipe(): %s\n", strerror(errno));
		return(-1);
	}
	
	pid = fork();
	
	if (pid == -1) {
		DEBUG("exec_pager(): fork(): %s\n", strerror(errno));
		close(p[0]);
		close(p[1]);
		return(-1);
	}
	
	// parent
	if (pid != 0) {
		// close pipe's read end
		close(p[0]);
		
		// move stdout out of the way
		fflush(stdout);
		ofd = dup(1);
		
		if (ofd < 0) {
			DEBUG("dup(stdout): %s\n", strerror(errno));
			close(p[1]);
			return(-1);
		}
		
		// move pipe's write end into stdout
		if (dup2(p[1], 1) != 1) {
			DEBUG("dup2(pipe_w, stdout): %s\n", strerror(errno));
			close(ofd);
			close(p[1]);
			return(-1);
		}
		
		if (p[1] > 1)
			close(p[1]);

		// return
		pg->pid = pid;
		pg->old_out = ofd;
		return(0);
	}
	
	// child
	
	// close pipe's write end
	close(p[1]);

	// move pipe's read end into stdin
	if (dup2(p[0], 0) != 0) {
		DEBUG("dup2(pipe_r, stdin): %s\n", strerror(errno));
		exit(1);
	}
	
	if (p[0] > 0)
		close(p[0]);
	
	// lookup pager
	// -F = --quit-if-one-screen
	// -X = --no-init: leave content onscreen on exit
	// -R = --RAW-CONTROL_CHARS: display colors correctly
	// -S = --chop-long-lines
	if (setenv("LESS", "FRS", 0)) {
		DEBUG("setenv(): %s\n", strerror(errno));
		errno = 0;
	}
	
	pager = getenv("PAGER");
	
	if (!pager)
		pager = "less";
	
	argv[0] = pager;
	argv[1] = NULL;

	// execute
	execvp(pager, argv);
	
	DEBUG("%s: %s\n", pager, strerror(errno));
	exit(1);
}


// **************************************************************************
//
// **************************************************************************
int exec_pager_wait(exec_pager *pg)
{
	pid_t r;
	int st, x;
	
	if (pg->pid == 0)
		return(0);
	
	x = 0;
	
	// close pipe's write end
	fflush(stdout);
	close(1);
	
	r = waitpid(pg->pid, &st, 0);
	
	if (r != pg->pid) {
		DEBUG("exec_pager_wait(): waitpid(): %s\n", strerror(errno));
		x = -1;
	} else if (!WIFEXITED(st)) {
		DEBUG("exec_pager_wait(): waitpid(): no exit status\n");
		x = -1;
	} else if (WEXITSTATUS(st) != 0) {
		DEBUG("exec_pager_wait(): waitpid(): exit status %d\n",
			WEXITSTATUS(st));
		x = -1;
	}
	
	if (pg->old_out >= 0) {
		if (dup2(pg->old_out, 1) != 1) {
			DEBUG("exec_pager_wait(): dup2(): %s\n",
				strerror(errno));
			x = -1;
		}
		
		close(pg->old_out);
	}
	
	pg->pid = 0;
	pg->old_out = -1;
	return(x);
}

// **************************************************************************
//
// **************************************************************************
int exec_data(char **argv, char *title, char *data)
{
	int p[2];
	size_t len;
	pid_t child, r;
	int status;
	
	if (pipe(p)) {
		DEBUG("exec_data(%s, %s): pipe(): %s\n",
			argv[0], title, strerror(errno));
		return(-1);
	}
	
	child = fork();
	
	// error
	if (child == -1) {
		DEBUG("exec_data(%s, %s): fork(): %s\n",
			argv[0], title, strerror(errno));
		close(p[0]);
		close(p[1]);
		return(-1);
	}
	
	// parent
	if (child != 0) {
		close(p[0]);
		
		len = strlen(data);

		if (block_write(p[1], data, len)) {
			close(p[1]);
			return(-1);
		}
		
		close(p[1]);
		
		if (waitpid(child, &status, 0) != child) {
			DEBUG("exec_data(%s, %s): waitpid(): %s\n",
				argv[0], title, strerror(errno));
			return(-1);
		}
		
		if (!WIFEXITED(status)) {
			DEBUG("exec_data(%s, %s): %s exited abnormally\n",
				argv[0], title, argv[0]);
			return(-1);
		}
		
		r = WEXITSTATUS(status);
		if (r != 0) {
			DEBUG("exec_data(%s, %s): %s returned %d\n",
				argv[0], title, argv[0], r);
			return(-1);
		}
		
		return(0);
	}

	// child

	// close pipe's write end
	close(p[1]);

	// move pipe's read end into stdin
	if (dup2(p[0], 0) != 0) {
		DEBUG("exec_data(%s, %s): dup2(pipe_read, stdin): %s\n",
			argv[0], title, strerror(errno));
		exit(1);
	}
	
	if (p[0] > 0)
		close(p[0]);
	
	// execute
	execvp(argv[0], argv);
	
	DEBUG("exec_data(%s, %s): %s: %s\n",
		argv[0], title, argv[0], strerror(errno));

	exit(1);
}



// **************************************************************************
//
// **************************************************************************
