/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <limits.h>
#include <stdlib.h>
#include <regex.h>
#include "pattern.h"


// ***************************************************************************
// 
// ***************************************************************************
static int pattern_range(int *lb, int *ub, char *s)
{
	char *t;
	char *e;
	
	*lb = strtol(s, &e, 0);
	
	if (e == s)
		return -1;

	if (e[0] == 0) {
		*ub = *lb;
		return 0;
	}
	
	if (e[0] != '-')
		return -1;
	
	if (e[1] == 0) {
		*ub = INT_MAX;
		return 0;
	}
	
	t = e + 1;
	
	*ub = strtol(t, &e, 0);
	
	if (*e != 0)
		return -1;
	
	return 0;
}

// ***************************************************************************
// 
// ***************************************************************************
#define PAT_OPTIONS	0x00ff
#define PAT_ALL		0x0100
#define PAT_NONE	0x0200
#define PAT_RANGE	0x0400
#define PAT_EXACT	0x0800
#define PAT_REGEX	0x1000
int pattern_init(pattern *pat, char *match, int flags)
{
	memset(pat, 0, sizeof(pattern));
	
	pat->flags = flags & PAT_OPTIONS;
	
	if ((match == NULL) || (match[0] == 0)) {
		pat->flags |= (flags & PATTERN_EMPTY_NONE)
			? PAT_NONE : PAT_ALL;
		return 0;
	}
	
	if (match[0] == '~') {
		if (regcomp(&pat->regex, match + 1, REG_EXTENDED | REG_NOSUB))
			return -1;

		pat->flags |= PAT_REGEX;
		return 0;
	}
	
	if ((flags & PATTERN_RANGES)
	&&  (pattern_range(&pat->id0, &pat->id1, match) == 0)) {
		pat->flags |= PAT_RANGE;
		return 0;
	}
	
	pat->match = match;
	pat->flags |= PAT_EXACT;
	
	return 0;
}

void pattern_clear(pattern *pat)
{
	if (pat->flags & PAT_REGEX) {
		regfree(&pat->regex);
	}
}

int pattern_matches(pattern *pat, int id, char *str)
{
	if (str == NULL)
		str = "";
	
	if (pat->flags & PAT_NONE)
		return 0;
	
	if (pat->flags & PAT_ALL)
		return 1;
	
	if (pat->flags & PAT_RANGE)
		return ((id >= pat->id0) && (id <= pat->id1));
	
	if (pat->flags & PAT_EXACT)
		return (strcmp(pat->match, str) == 0);
	
	if (pat->flags & PAT_REGEX)
		return (regexec(&pat->regex, str, 0, NULL, 0) == 0);
	
	return 0;
}


// ***************************************************************************
// 
// ***************************************************************************
